// Copyright 2010 Rich Townsend <townsend@astro.wisc.edu>
//
// This file is part of CULSP.
//
// CULSP is free software: you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// CULSP is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with CULSP. If not, see <http://www.gnu.org/licenses/>.

// Includes

#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <argtable2.h>
#include <omp.h>

#include "periodogram.h"

#define TWOPI 6.2831853071796f

// Forward declarations

void initialize (int, char **, char **, char **, float *, float *);
void eval_LS_periodogram (float *, float *, float *, int, float, int);

// Main program

int
main( int argc, char** argv) 
{

  char *filename_in;
  char *filename_out;
  float F_over;
  float F_high;

  int N_t;
  int N_f;
 
  float *t;
  float *X;

  float df;

  float *P;

  // Initialize

  initialize(argc, argv, &filename_in, &filename_out, &F_over, &F_high);

  // Read the light curve

  read_light_curve(filename_in, &N_t, &t, &X);

  // Set up the frequency parameters

  set_frequency_params(N_t, t, F_over, F_high, &N_f, &df);

  // Allocate space for the periodogram

  P = (float *) malloc(N_f*sizeof(float));

  // Start the timer

  double time_a = get_time();
		 
  // Evaluate the Lomb-Scargle periodogram
 
  eval_LS_periodogram(t, X, P, N_t, df, N_f);

  // Stop the timer

  double time_b = get_time();
  printf( "Processing time: %16.3f (ms)\n", (time_b-time_a)*1000);

  // Write the data to file

  write_periodogram(filename_out, N_f, df, P);

  // Free up space

  free(P);
  free(X);
  free(t);

  // Finish

}


////
// Initialization
////

void
initialize (int argc, char **argv, char **filename_in, char **filename_out, 
	    float *F_over, float *F_high)
{

  // Set up the argtable structs

  struct arg_file *in = arg_file1(NULL, "in", "<filename_in>", "input file");
  struct arg_file *out = arg_file1(NULL, "out", "<filename_out>", "output file");
  struct arg_dbl *over = arg_dbl0(NULL, "over", "<F_over>", "oversample factor");
  struct arg_dbl *high = arg_dbl0(NULL, "high", "<F_high>", "highest-frequency factor");
  struct arg_end *end = arg_end(20);
  void *argtable[] = {in,out,over,high,end};

  // Parse the command line

  int n_error = arg_parse(argc, argv, argtable);

  if (n_error == 0) {

    *filename_in = (char *) malloc(strlen(in->filename[0])+1);
    strcpy(*filename_in, in->filename[0]);

    *filename_out = (char *) malloc(strlen(out->filename[0])+1);
    strcpy(*filename_out, out->filename[0]);

    *F_over = over->count == 1 ? (float) over->dval[0] : 1.f;
    *F_high = high->count == 1 ? (float) high->dval[0] : 1.f;

  }
  else {

    printf("Syntax: %s", argv[0]);
    arg_print_syntax(stdout, argtable, "\n");

    exit(EXIT_FAILURE);

  }

  // Finish

}


////
// Periodogram evaluation
////

void
eval_LS_periodogram (float *t, float *X, float *P, int N_t, float df, int N_f)
{

  printf("Launching calculations...\n");

  // Loop through the frequencies

  #pragma omp parallel for
  for(int i = 0; i < N_f; i++) {

    // Calculate the frequency

    float f = (i+1)*df;

    // Calculate the various sums

    float XC = 0.f;
    float XS = 0.f;
    float CC = 0.f;
    float CS = 0.f;

    float XC_chunk = 0.f;
    float XS_chunk = 0.f;
    float CC_chunk = 0.f;
    float CS_chunk = 0.f;

    int j;

    for(j = 0; j < N_t-BLOCK_SIZE; j += BLOCK_SIZE) {

      // Update the sums

      #pragma unroll
      for(int k = 0; k < BLOCK_SIZE; k++) {

	// Range reduction

	float ft = f*t[j+k];
	ft -= (int) ft;

	float c = cosf(TWOPI*ft);
	float s = sinf(TWOPI*ft);

	XC_chunk += X[j+k]*c;
	XS_chunk += X[j+k]*s;
	CC_chunk += c*c;
	CS_chunk += c*s;

      }

      XC += XC_chunk;
      XS += XS_chunk;
      CC += CC_chunk;
      CS += CS_chunk;

      XC_chunk = 0.f;
      XS_chunk = 0.f;
      CC_chunk = 0.f;
      CS_chunk = 0.f;

    }

    // Handle the final chunk

    for(int k = 0; k < N_t-j; k++) {

      // Range reduction

      float ft = f*t[j+k];
      ft -= (int) ft;

      float c = cosf(TWOPI*ft);
      float s = sinf(TWOPI*ft);

      XC_chunk += X[j+k]*c;
      XS_chunk += X[j+k]*s;
      CC_chunk += c*c;
      CS_chunk += c*s;

    }

    XC += XC_chunk;
    XS += XS_chunk;
    CC += CC_chunk;
    CS += CS_chunk;

    float SS = N_t - CC;

  // Calculate the tau terms

    float omega_tau = 0.5f*atan2f(2*CS, CC-SS);

    float st = sinf(omega_tau);
    float ct = cosf(omega_tau);

    // Calculate P

    P[i] = 
      0.5f*((ct*XC + st*XS)*(ct*XC + st*XS)/
	    (ct*ct*CC + 2*ct*st*CS + st*st*SS) + 
	    (ct*XS - st*XC)*(ct*XS - st*XC)/
	    (ct*ct*SS - 2*ct*st*CS + st*st*CC));

  }

  // Finish

  printf("Completed!\n");

}
